"""Provides different backends for a project.

The three main kinds of backends are:
- InMemoryBackend - this is a backend that does not have persistent storage of
  the data, once the project is unloaded the data is lost. If the data created
  is needed it needs to be exported out. This is commonly used for tests.
- ExistingBackend - the project is backed by a project that exists in an
  existing (running) application. This is the commonly used by users of this
  SDK.
- NewBackend - a new backend is created to host the project. A use case of this
  would be reading/writing to a Maptek Database that isn't opened in an
  application or the SDK is being used to create an application.

The alternative name would have been subsystem so as not to confuse these
backends with the concept of backends within the C++ side. For example, the
different project backends can be Maptek Database, Maptek Object, Vulcan
DGD/TEK Isis or a Vulcan project directory.
"""
###############################################################################
#
# (C) Copyright 2023, Maptek Pty Ltd. All rights reserved.
#
###############################################################################

from __future__ import annotations

import ctypes
import os

from . import account
from .options import ProjectOptions, ProjectBackendType
from .mcp import (ExistingMcpdInstance, McpdConnection, McpdDisconnectError,
                  connect_to_mcpd, find_mdf_hosts)
from .. import capi
from ..project.errors import ProjectConnectionFailureError


class Backend:
  """An abstract class for representing the different backends for projects.

  See the derived classes for specifics.

  An alternate name for this would be ProjectSubsystem.
  """

  def __init__(self, log):
    self.broker_session = None
    self.index = None
    self.licence = ''
    self.log = log

  def acquire_licence(self,
                      broker_connector_path: str | None,
                      connection_parameters: dict | None) -> str:
    """Acquire a licence for using the given backend.

    This uses the Licence DLL to check the supported licence format.

    Parameters
    ----------
    broker_connector_path
      The path to where the Maptek Account Broker assemblies can be found.
      If None, the assemblies bundled with the Python SDK are used.
    connection_parameters
      Provides parameters sent to the Maptek Account broker.

    Returns
    -------
    str
      The acquired licence.
    """
    # First, try with an anonymous session. This allows the use of borrowed
    # licences when the user isn't logged into Maptek Account.
    #
    # Secondly, if that that fails then we try non-anonymous so the user is
    # prompted to log-in. Unless the caller has been explicit and said they
    # only want an anonymous session.
    required_parameters = {
        'AnonymousSession': True,
        'MaptekAccountUserName': '',
        'MaptekAccountAuthKey': '',
        'ApiToken': '',
      }
    if connection_parameters:
      force_parameters = connection_parameters.copy()
      force_parameters.update(required_parameters)
    else:
      force_parameters = required_parameters

    supported_licence_format = capi.License().supported_licence_format()
    broker_error : ValueError | None = None
    for try_parameters in [force_parameters, connection_parameters]:
      try:
        self.broker_session = account.connect_to_maptek_account_broker(
          broker_connector_path, try_parameters)

        with self.broker_session.acquire_extend_licence(
            supported_licence_format) as licence:
          self.log.info('Acquired licence for Maptek Extend')
          self.licence = licence.license_string
          return licence.license_string
      except ValueError as error:
        # That failed, but we can try again this time without requiring it
        # to be anonymous to allow the user to login.
        if self.broker_session:
          self.broker_session.disconnect()

        # There are two options for how to log the message here. One is to be
        # generic to match how this code is written (which is rather generic)
        # so it would be future-proof, for example "Initial attempt to acquire
        # a licence has failed, trying again.". Two is to actually say the
        # intent at the time which was "Failed to find a borrowed licence.
        # Trying a live licence."
        self.log.info('Failed to find a borrowed licence. Trying a live '
                      'licence.')

        # However it is possible the caller has explicitly said be
        # anonymous in which case this won't make any difference.
        if connection_parameters and connection_parameters.get(
            'AnonymousSession', False):
          raise

        # Capture the error so it can throw if it was the last one.
        broker_error = error

    raise broker_error # type: ignore Variable set in except.

  def connect_to_project(self) -> None:
    """Connects to the project.

    In older versions of the applications, this would need to be called
    per-thread.

    This populates self.index with the index of the backend.

    Raises
    ------
    ProjectConnectionFailureError
      If connecting to the project fails.
    """
    raise NotImplementedError("Derived classes should implement this method")

  def disconnect_from_project(self) -> None:
    """Disconnect from the project.

    self.index is reset back to to None.

    After disconnecting from the project, it will then disconnect from the
    Maptek Account broker.

    Calling this more than once is acceptable and won't raise an error.
    """
    try:
      if self.index is not None:
        self._disconnect_from_project()
    finally:
      self.index = None
      if self.broker_session:
        self.broker_session.disconnect()
        self.broker_session = None

  def _disconnect_from_project(self) -> None:
    """Handle the backend specific disconnection from the project.

    The specific backend should handle any clean-up for the work done during
    connect_to_project().

    Implementations need to override this function.
    The overridden function should not call this function.
    """
    raise NotImplementedError("Derived classes should implement this method")


class InMemoryBackend(Backend):
  """A backend for the project that is stored in-memory only.

  The data is not written out to a file.  If any data should persist after
  the project is closed, it will need to be exported.
  """

  def connect_to_project(self) -> None:
    """Connects to the DataEngine.

    This creates a DataEngine in the current process.
    """
    # No executables or servers will be used. This means no mcpd or
    # backendServer will be launched or used.
    connected = capi.DataEngine().CreateLocal()
    if not connected:
      last_error = capi.DataEngine().ErrorMessage().decode(
        "utf-8")
      error_message = ("There was an error while creating"
        f" memory-only project ({last_error})")
      self.log.critical(error_message)
      raise ProjectConnectionFailureError(error_message)

    self.index = 0
    self.log.info("Created memory-only DataEngine")

  def _disconnect_from_project(self):
    """Disconnects from the DataEngine."""
    # A project backed onto only a DataEngine isn't expected to have any other
    # clients. Its primary use is for tests and scripts which need to
    # open a DataEngine, create/import and save results then close it, i.e.
    # not in multi-process situations where the lifetime of the processes
    # are unknown.
    self.log.info("Disconnecting from an in-memory project.")
    close_backends_if_last_client = False
    capi.DataEngine().Disconnect(close_backends_if_last_client)


class ExistingBackend(Backend):
  """A backend for the project that is hosted by another application.

  The data is typically stored in to a Maptek Database (Under a .maptekdb
  directory).

  The application is expected to have the backend server running so this
  should not need to start one.
  """

  def __init__(self, log, mcpd_instance):
    super().__init__(log)
    self.mcpd_instance = mcpd_instance

    # Populated by connect_to_project() or use_existing_connection()
    self.mcp_connection: McpdConnection | None = None

    # If an existing connection is given then it is not the responsibility
    # of this class to disconnect from it.
    self.connection_our_responsibility = True

  @classmethod
  def find_application_to_connect_to(cls, log) -> ExistingMcpdInstance | None:
    """Find a suitable application to connect to.

    This first checks environment variables set by the Maptek Workbench.
    It does this because when it runs a Python script it can recommend an
    application to connect to, such as the current active application.

    If those variables are not set then it looks at running application or
    more specifically the MCPD running as part of those applications.

    Parameters
    ----------
    log
      The logger for writing messages.
      If None then no logging is performed.

    Returns
    -------
    ExistingMcpdInstance
      The MCPD instance to connect to.
    None
      If there was no host application found.
    """
    dll_path = os.environ.get("SDK_OVERWRITE_BIN_PATH", "")
    socket_path = os.environ.get("SDK_OVERWRITE_MCP_PATH", "")
    if dll_path and socket_path:
      return ExistingMcpdInstance(-1, dll_path, socket_path)

    instances = find_mdf_hosts(log)
    if instances:
      return instances[0]

    return None

  def use_existing_connection(self, connection: McpdConnection):
    """Use an existing connection to the MCPD.

    This means when connecting to the project, it won't need to connect
    on its own.

    The backend does not take responsibility for closing the connection,
    that remains the callers responsibility.
    """
    self.mcp_connection = connection
    self.connection_our_responsibility = True

  def connect_to_project(self) -> None:
    # Connect to the MCPD if an existing McpdConnection was not supplied.
    if not self.mcp_connection:
      self.mcp_connection = connect_to_mcpd(
      specific_mcpd=self.mcpd_instance,
      sdk_licence=self.licence,
      )

    # Connecting to an existing DataEngine session.
    create_new_session = False
    connected = capi.DataEngine().Connect(create_new_session)
    if not connected:
      last_error = capi.DataEngine().ErrorMessage().decode("utf-8")
      error_message = (
        f"There was an error connecting to the database ({last_error})")
      self.log.critical(error_message)
      raise ProjectConnectionFailureError(error_message)

    # The backend is managed by an existing application; what kind of backend
    # is not provided when connecting to it, however it can be queried from
    # the root object.
    backend_index = ctypes.c_uint16()
    if not capi.DataEngine().ObjectBackEnd(capi.DataEngine().RootContainer(),
                                            ctypes.byref(backend_index)):
      last_error = capi.DataEngine().ErrorMessage().decode("utf-8")
      error_message = (
        "Unable to determine the backend used by the application "
        f"({last_error})")
      self.log.critical(error_message)
      raise ProjectConnectionFailureError(error_message)

    self.index = backend_index

  def _disconnect_from_project(self) -> None:
    # The backend for the project is hosted by another application, we simply
    # need to disconnect from it. However if our connection was keeping the
    # backend alive we need to close it if we are the last one.
    self.log.info("Disconnecting from a project opened by an another "
                  "application.")
    close_backends_if_last_client = True
    capi.DataEngine().Disconnect(close_backends_if_last_client)

    if self.connection_our_responsibility:
      try:
        self.mcp_connection.disconnect()
      except McpdDisconnectError:
        # The exception is already logged, so it is safe to ignore it.
        pass


class NewBackend(Backend):
  """A backend for the project that is created by Python.

  The data will be stored in a Maptek Database (Under a .maptekdb
  directory).
  """

  def __init__(self, log, mcpd_instance, options: ProjectOptions):
    super().__init__(log)
    self.options = options

    if isinstance(mcpd_instance, McpdConnection):
      raise TypeError('mcpd_instance should not be a connection. To use an '
                      'existing connection use_existing_connection() should '
                      'be called instead.')

    self.mcpd_instance = mcpd_instance

    # Populated by connect_to_project().
    self.mcp_connection: McpdConnection | None = None

    # If an existing connection is given then it is not the responsibility
    # of this class to disconnect from it.
    self.connection_our_responsibility = True

  def use_existing_connection(self, connection: McpdConnection):
    """Use an existing connection to the MCPD.

    This means when connecting to the project, it won't need to connect
    on its own.

    The backend does not take responsibility for closing the connection,
    that remains the callers responsibility.
    """
    self.mcp_connection = connection
    self.connection_our_responsibility = False

  def connect_to_project(self) -> None:
    # This expects that a connection is already made to the mcpd.exe and
    # it will be able to launch the backendServer.

    # Connect to the MCPD if an existing McpdConnection was not supplied.
    if not self.mcp_connection:

      if not self.mcpd_instance:
        # In the future, this may be where an MCPD is spawned. To achieve this
        # the licence acquired should have the necessary feature codes for
        # the MCPD.
        raise ValueError('A connection to the MCPD or the details for MCPD '
                         'instance is required')

      self.mcp_connection = connect_to_mcpd(
        specific_mcpd=self.mcpd_instance,
        sdk_licence=self.licence,
        )

    # Start a backend server.
    if 'MDF_BIN' not in os.environ:
      expected_path = os.path.join(self.options.dll_path, 'mcpd.exe')
      if os.path.exists(expected_path):
        os.environ['MDF_BIN'] = expected_path
      else:
        # This accounts for a situation that is unique to the
        # under-development versions of the software.
        if os.path.exists(os.path.join(self.options.dll_path, '..', 'bin',
                                        'mcpd.exe')):
          os.environ['MDF_BIN'] = os.path.join(
            self.options.dll_path, '..', 'bin')
        else:
          error_message = "Failed to register backend server - environment " + \
            "variable MDF_BIN not set."
          self.log.critical(error_message)
          raise ProjectConnectionFailureError(error_message)

    # Ideally, the absolute path would be provided to the backendServer,
    # however the underlying function doesn't handle spaces. Using an
    # environment variable allows the given string to have no spaces but
    # for the path to contain spaces when it is expanded.
    backend_registered = self.mcp_connection.register_server(
      '$MDF_BIN/backendServer')
    if not backend_registered:
      error_message = "Failed to register backend server"
      self.log.critical(error_message)
      raise ProjectConnectionFailureError(error_message)

    # Ensure there are no stale lock files.
    if not capi.DataEngine().DeleteStaleLockFile(
        self.options.project_path.encode('utf-8')):
      last_error = capi.DataEngine().ErrorMessage().decode("utf-8")
      self.log.error(
        "There was a problem ensuring no stale lock files "
        "had been left in the project. Error message: %s", last_error)

      # Try sharing the project?
      self.options.backend_type = ProjectBackendType.SHARED
      self.log.warning('Attempting to share the project "%s"',
                        self.options.project_path)

    # Create or open project.
    self.index = capi.DataEngine().OpenProject(
      self.options.project_path.encode('utf-8'),
      self.options.open_mode,
      self.options.access_mode,
      self.options.backend_type,
      self.options.proj_units)
    if self.index == 0:
      last_error = capi.DataEngine().ErrorMessage().decode("utf-8")
      error_message = ("There was a problem using the requested database "
                       "load or creation settings. "
                       f"Error message: {last_error}")
      self.log.critical(error_message)
      raise ProjectConnectionFailureError(error_message)

    # Connecting to an existing DataEngine session (the session created by the
    # backendServer when the project was opened).
    create_new_session = False
    connected = capi.DataEngine().Connect(create_new_session)
    if not connected:
      last_error = capi.DataEngine().ErrorMessage().decode("utf-8")
      error_message = (
        f"There was an error connecting to the database ({last_error})")
      self.log.critical(error_message)
      raise ProjectConnectionFailureError(error_message)

  def _disconnect_from_project(self) -> None:
    if self.index is None:
      return  # No project to disconnect from.
    if self.index == 0:
      raise ValueError("Unexpected backend index")

    self.log.info("Closing project with backend index: %s", self.index)
    capi.DataEngine().CloseProject(self.index)

    if self.connection_our_responsibility:
      try:
        self.mcp_connection.disconnect()
      except McpdDisconnectError:
        # The exception is already logged, so it is safe to ignore it.
        pass

    self.mcp_connection = None
